(function (global, factory) {
typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
typeof define === 'function' && define.amd ? define(['exports'], factory) :
(global = global || self, factory(global.machines = {}));
}(this, function (exports) { 'use strict';

/**
 * Compares a current state value to a transition name, using a state machine (Object) where all the possible states also have transition names registered to each one (which link to new states).
 * Invoking this function will make that calculation and always return to you what the next state should be (which may be un-changed).
 * This function is used by `createMachine()`.
 *
 * @function
 * @name getNextState
 * @sig String -> String -> {k: v} -> String
 * @param {string} currentState The current state - which matches one of the keys in the 'stateMachine'
 * @param {string} transitionName A string value that will cause a state change
 * _if_ it matches the key name of one (or more) of the sub-objects in the 'stateMachine'
 * @param {Object<string, Object<string, function>>} stateMachine An Object of (only) Objects. Inside each of
 * those Objects all values are Strings which correspond to keys on the outer Object.
 * @returns {string} The next state (which might not have changed from the current state)
 */
function getNextState(currentState, transitionName, stateMachine) {
  if (stateMachine === void 0) {
    stateMachine = {};
  }

  var registeredTransitions = stateMachine[currentState];
  if (!registeredTransitions) return currentState;
  var possibleNewState = registeredTransitions[transitionName];
  return possibleNewState || currentState;
}

var getNextState_1 = getNextState;

/**
 * Checks if a given value is an Object {}
 *
 * @function
 * @name _isObject
 * @private
 * @sig * -> Boolean
 * @param {*} val A value which may possibly be an object
 * @returns {boolean} Whether or not the given values is an object
 */
function _isObject(val) {
  return typeof val === 'object' && val && typeof val.length === 'undefined';
}
/**
 * Checks a given object to make sure that all its values are also objects
 *
 * @function
 * @name _allValuesAreObjects
 * @private
 * @sig {k: v} -> Boolean
 * @param {Object<string, Object<string, string>>} obj An Object of (only) Objects. Inside each of those Objects all values are Strings which correspond to keys on the outer Object.
 * @returns {boolean} Whether or not the object contains values which are all also objects
 */


function _allValuesAreObjects(obj) {
  return Object.keys(obj || {}).every(function (k) {
    return _isObject(obj[k]);
  });
}
/**
 * Checks that each value the transitions lead to on a state machine are also states on the state machine
 *
 * @function
 * @name _allValuesFromTransitionsAreStates
 * @private
 * @sig {k: v} -> Boolean
 * @param {Object<string, Object<string, string>>} obj An Object of (only) Objects. Inside each of those Objects all values are Strings which correspond to keys on the outer Object.
 * @returns {boolean} Whether or not all values that the transitions lead to are also states on the state machine
 */


function _allValuesFromTransitionsAreStates(obj) {
  if (obj === void 0) {
    obj = {};
  }

  var isInMachine = function isInMachine(transitionState) {
    return Object.prototype.hasOwnProperty.call(obj, transitionState);
  };

  return Object.keys(obj).map(function (state) {
    return obj[state];
  }).every(function (transitions) {
    return Object.keys(transitions).every(function (tstate) {
      return isInMachine(tstate);
    });
  });
}
/**
 * Checks a given state machine to make sure it is valid, and throws an error if not
 *
 * @function
 * @name validateStateMachine
 * @sig {k: v} -> undefined
 * @throws {TypeError} If the state machine is missing or not an object
 * @throws {TypeError} If the state machine has any values which are not objects
 * @throws {TypeError} If the state machine's "sub-objects" has values that are not keys on the root of the object
 *
 * @param {Object<string, Object<string, string>>} stateMachine An Object of (only) Objects. Inside each of those Objects all values are Strings which correspond to keys on the outer Object.
 */


function validateStateMachine(stateMachine) {
  if (!_isObject(stateMachine)) {
    throw new TypeError('A state machines must be an Object');
  }

  if (!_allValuesAreObjects(stateMachine)) {
    throw new TypeError('A state machine must be an Object of (only) Objects');
  }

  if (_allValuesFromTransitionsAreStates(stateMachine)) {
    throw new TypeError('All the registered transitions (for each possible state) in a machine must lead to another state in the machine');
  }
}
/**
 * Gets the first state in a given state machine
 *
 * @function
 * @name getFirstState
 * @sig {k: v} -> String -> String
 * @param {Object<string, Object<string, string>>} stateMachine An Object of (only) Objects. Inside each of those Objects all values are Strings which correspond to keys on the outer Object.
 * @param {string} initialState The state at which the machine should be initialized (will default to the first key name in the state machine)
 * @returns {string} The state value at which the machine should start
 */


function getFirstState(stateMachine, initialState) {
  var currentState = initialState == null ? Object.keys(stateMachine)[0] : initialState;

  if (typeof currentState !== 'string' || !Object.prototype.hasOwnProperty.call(stateMachine, currentState)) {
    throw new Error("The \"initialState\" must be one of the keys on the \"stateMachine\" itself " + currentState);
  }

  return currentState;
}

var _internal = {
  getFirstState: getFirstState,
  validateStateMachine: validateStateMachine,
  _isObject: _isObject,
  _allValuesAreObjects: _allValuesAreObjects,
  _allValuesFromTransitionsAreStates: _allValuesFromTransitionsAreStates
};

var getFirstState$1 = _internal.getFirstState,
    validateStateMachine$1 = _internal.validateStateMachine;
/**
 * Creates a state machine - which is managed inside a closure - that will only transition if name of the transition you pass into it is among those registered for the current state.
 * You pass in a state machine Object and you'll have a function returned to you that can repeatedly receive a transition name and always return to you the next state (if it even changes at all).
 *
 * The current state will always match one of the keys in the state machine object.
 * Associated with those keys are each Objects that hold the possible transitions that can move that state to another one.
 *
 * So each sub-object in the state machine holds transition names as keys and _states-to-transition-to_ as values.
 * The curent state can only change if a given transition's name is among those registered for the current state.
 * Only registered transition names can produce changes to the current state.
 *
 * If you use this with Redux then you'll probably want to make the transition names match the names of Redux Action Types.
 * That way your state machine can be used in a reducer (or a piece of Redux middleware) to alter a single prop on your Redux store (representing that machine's current state).
 *
 * @function
 * @name createMachine
 * @sig {k: v} -> String -> (String -> String)
 * @throws {TypeError} If the state machine is missing or not an object
 * @throws {TypeError} If the state machine has any values which are not objects
 * @throws {TypeError} If the state machine's "sub-objects" has values that are not keys on the root of the object
 * @throws {Error} If the initialState is not one of the state values on the state machine
 * @param {Object<string, Object<string, string>>} stateMachine An Object of (only) Objects. Inside each of those Objects all values are Strings which correspond to keys on the outer Object.
 * @param {string} initialState The state at which the machine should be initialized (will default to the first key name in the state machine)
 * @returns {function} A getNextState() function that is ready to receive your transition name and return the next state (which may or not be different).
 * @example
 * const authMachine = {
 *   initial: {
 *     ATTEMPT_LOGIN: 'inProgress'
 *   },
 *   inProgress: {
 *     LOGIN_ERROR: 'error',
 *     LOGOUT_ERROR: 'error',
 *     LOGIN_SUCCESSFUL: 'loggedIn',
 *     LOGOUT_SUCCESSFUL: 'loggedOut'
 *   },
 *   loggedIn: {
 *     ATTEMPT_LOGOUT: 'inProgress'
 *   },
 *   loggedOut: {
 *     ATTEMPT_LOGIN: 'inProgress'
 *   },
 *   error: {
 *     ATTEMPT_LOGIN: 'inProgress',
 *     CLEAR_ERROR: 'loggedOut'
 *   }
 * }
 *
 * const getNextAuthState = createMachine(authMachine, 'initial')
 *
 * getNextAuthState('LOGIN_SUCCESSFUL')
 * // initial - Can't transition there yet
 * getNextAuthState('CLEAR_ERROR')
 * // initial - Also doesn't affect state
 * getNextAuthState('ATTEMPT_LOGIN')
 * // inProgress - Now we have a login attempt, which changes state
 * getNextAuthState('LOGIN_SUCCESSFUL')
 * // loggedIn - Advances from that pending state to logged-in
 * getNextAuthState('ATTEMPT_LOGOUT')
 * // inProgress - Attempting to log out
 * getNextAuthState('LOGOUT_ERROR')
 * // error - Now we're in an error state
 * getNextAuthState('CLEAR_ERROR')
 * // loggedOut - Finishes logout process
 */

function createMachine(stateMachine, initialState) {
  validateStateMachine$1(stateMachine);
  var currentState = getFirstState$1(stateMachine, initialState);
  /**
   * A simple function that retrieves the current state or optionally advances to the next state determined by a given transition value
   *
   * @function
   * @name nextState
   * @sig String -> String
   * @param {string} [transition] One of the transition names registered for the current state
   * @returns {string} The current state (one of the root-level keys from the state machine
   */

  return function nextState(transition) {
    currentState = getNextState_1(currentState, transition, stateMachine);
    return currentState;
  };
}

var createMachine_1 = createMachine;

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  subClass.__proto__ = superClass;
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function isNativeReflectConstruct() {
  if (typeof Reflect === "undefined" || !Reflect.construct) return false;
  if (Reflect.construct.sham) return false;
  if (typeof Proxy === "function") return true;

  try {
    Date.prototype.toString.call(Reflect.construct(Date, [], function () {}));
    return true;
  } catch (e) {
    return false;
  }
}

function _construct(Parent, args, Class) {
  if (isNativeReflectConstruct()) {
    _construct = Reflect.construct;
  } else {
    _construct = function _construct(Parent, args, Class) {
      var a = [null];
      a.push.apply(a, args);
      var Constructor = Function.bind.apply(Parent, a);
      var instance = new Constructor();
      if (Class) _setPrototypeOf(instance, Class.prototype);
      return instance;
    };
  }

  return _construct.apply(null, arguments);
}

function _isNativeFunction(fn) {
  return Function.toString.call(fn).indexOf("[native code]") !== -1;
}

function _wrapNativeSuper(Class) {
  var _cache = typeof Map === "function" ? new Map() : undefined;

  _wrapNativeSuper = function _wrapNativeSuper(Class) {
    if (Class === null || !_isNativeFunction(Class)) return Class;

    if (typeof Class !== "function") {
      throw new TypeError("Super expression must either be null or a function");
    }

    if (typeof _cache !== "undefined") {
      if (_cache.has(Class)) return _cache.get(Class);

      _cache.set(Class, Wrapper);
    }

    function Wrapper() {
      return _construct(Class, arguments, _getPrototypeOf(this).constructor);
    }

    Wrapper.prototype = Object.create(Class.prototype, {
      constructor: {
        value: Wrapper,
        enumerable: false,
        writable: true,
        configurable: true
      }
    });
    return _setPrototypeOf(Wrapper, Class);
  };

  return _wrapNativeSuper(Class);
}

var validateStateMachine$2 = _internal.validateStateMachine;
/**
 * Exends the base JavaScript base `Error`
 * @typedef {Object<string, any>} TransitionError
 * @property {string} name The name of the error (in this case "TransitionError")
 * @property {string} data.currentState The current state of the state machine at the time the error was encountered
 */

var TransitionError =
/*#__PURE__*/
function (_Error) {
  _inheritsLoose(TransitionError, _Error);

  function TransitionError(message, state, additionalData) {
    var _this;

    if (additionalData === void 0) {
      additionalData = {};
    }

    _this = _Error.call(this, message) || this;
    _this.name = _this.constructor.name;
    _this.data = _objectSpread({}, typeof additionalData === 'object' && additionalData.constructor.name === 'Object' ? additionalData : {}, {
      currentState: state
    });
    return _this;
  }

  return TransitionError;
}(_wrapNativeSuper(Error));
/**
 * Takes a given state machine and handler and returns a function which is ready to receive the starting "state" value _and_ any starting data, following through the entire sequence of async function until the state stops advancing.
 * Note, a handler will receive the initial data, the starting "state" value and any (optional) dependencies to place into the context for the async functions contained in the handler.
 * A handler doesn't have to be a function but can also be an object whose keys are possible "state" values and whose values are functions that receive the initial data, state machine, and context.
 *
 * @function
 * @name createTransitionHandler
 * @throws {TypeError} If a state machine is not provided
 * @throws {TypeError} If a handler function/object is not provided
 * @throws {TypeError} If the handler is not a function or an object whose keys do not correspond to keys in the state `machine`
 * @throws {TypeError} If the state machine has any values which are not objects
 * @throws {TypeError} If the state machine's "sub-objects" has values that are not keys on the root of the object
 * @param {function|Object<string, function>} handler A handler function or an object of individual handler function (the keys on the object must match possible "states" otherwise they'll never be called)
 * @param {Object<string, Object<string, string>>} machine An object whose keys are possible states and whose values are also objects (but whose keys are transition names and whose values are states they will advance to)
 * @returns {function} A wrapped handler function that is ready to receive the (1) initial data, (2) starting "state" value on the machine, and (3) an object containing any dependencies the handler may have
 */


function createTransitionHandler(handler, machine) {
  if (!machine) {
    throw new TypeError('A state machine is required but was not provided');
  }

  if (!handler) {
    throw new TypeError('A handler is required for the state machine transitions');
  }

  validateStateMachine$2(machine);
  var getStateHandler = typeof handler === 'function' ? function () {
    return handler;
  } : handler && typeof handler === 'object' && handler.constructor.name === 'Object' ? function (currentState) {
    if (typeof handler[currentState] !== 'function') {
      throw new TypeError("No handler was defined for the current state of '" + currentState + "'");
    }

    return handler[currentState];
  } : undefined;

  if (!getStateHandler) {
    throw new TypeError('handler must be a function or an object whose keys are state names and whose values are functions');
  }
  /**
   * A handler function that will advance from state to state - executing each piece of logic and/or async function the user defined, until the state can no longer advance and the current value or error is returned at that point.
   *
   * @function
   * @name wrappedTransitionHandler
   * @throws {TransitionError} If the handler encounters an unexpected exception
   * @param {Object<string, any>} initialData Any starting data for the handler
   * @param {string} initialState A starting "state" value which should be one of the possible values on the state machine
   * @param {Object<string, any>} [context] An optional object containing any dependencies of the handler (API clients, caches, etc.)
   * @returns {*} When the handler can no longer advance to another possible state, the value of the last function it executed is returned
   */


  return function wrappedTransitionHandler(initialData, initialState, context) {
    var nextState = createMachine_1(machine, initialState || 'initial');
    var currentState;
    var result = initialData;

    function next() {
      if (currentState !== nextState()) {
        currentState = nextState();
        var response = getStateHandler(currentState)(result, nextState, context);

        if (typeof response.then === 'function') {
          response.then(function (r) {
            result = r;
            return next();
          });
        } else {
          result = response;
          next();
        }
      }
    }

    try {
      next();

      if (result && /Error$/.test(result.constructor.name)) {
        throw result;
      }

      return result;
    } catch (err) {
      throw new TransitionError(err.message || err.toString(), nextState(), _objectSpread({}, err.data && typeof err.data === 'object' || {}, err.extensions && typeof err.extensions === 'object' || {}));
    }
  };
}

var createTransitionHandler_1 = createTransitionHandler;

var src = createMachine_1;
var getNextState_1$1 = getNextState_1;
var createTransitionHandler_1$1 = createTransitionHandler_1;
src.getNextState = getNextState_1$1;
src.createTransitionHandler = createTransitionHandler_1$1;

exports.createTransitionHandler = createTransitionHandler_1$1;
exports.default = src;
exports.getNextState = getNextState_1$1;

Object.defineProperty(exports, '__esModule', { value: true });

}));
